/*
 *  AMT KVM authentication functions.
 *
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <poll.h>
#include <unistd.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>
#include "net.h"
#include "auth.h"
#include "amtkvm.h"

struct redirinfo
{
	int major;
	int minor;
	int fwmajor;
	int fwminor;
	int iana;
	int length;
	struct in_addr addr;
	unsigned char data[255];
};

static int start_redirect(struct ctx *ctx,int persist,char *what,
	struct redirinfo *info)
{
	int length;
	unsigned char bfr[255];

	if(info)memset(info,0,sizeof(struct redirinfo));

	memset(bfr,0,sizeof(bfr));
	bfr[0]=0x10;
	if(persist)bfr[1]=0x01;
	memcpy(bfr+4,what,4);
	if(io(ctx,bfr,8,WRITE))return -1;

	if(io(ctx,bfr,13,READ))return -1;
	if(bfr[0]!=0x11)return -1;
	switch(bfr[1])
	{
	case 0x00:
		length=bfr[12];
		if(info)
		{
			info->major=bfr[4];
			info->minor=bfr[5];
			info->fwmajor=bfr[6];
			info->fwminor=bfr[7];
			info->iana=bfr[11];
			info->iana<<=8;
			info->iana|=bfr[10];
			info->iana<<=8;
			info->iana|=bfr[9];
			info->iana<<=8;
			info->iana|=bfr[8];
			info->length=length;
			if(io(ctx,info->data,length,READ))return -1;
		}
		else if(io(ctx,bfr,length,READ))return -1;
		return 0;

	case 0x02:
		if(info)memcpy(&info->addr,bfr+4,4);
		return -2;

	default:return -1;
	}
}

static int end_redirect(struct ctx *ctx)
{
	unsigned char bfr[4];

	memset(bfr,0,sizeof(bfr));
	bfr[0]=0x12;
	if(io(ctx,bfr,4,WRITE))return -1;

	return 0;
}

static int start_kvm(struct ctx *ctx)
{
	int status;
	unsigned char bfr[8];

	memset(bfr,0,sizeof(bfr));

	bfr[0]=0x40;
	if(io(ctx,bfr,8,WRITE))return -1;

	if(io(ctx,bfr,8,READ))return -1;
	if(bfr[0]!=0x41)return -1;
	status=bfr[7];
	status<<=8;
	status|=bfr[6];
	status<<=8;
	status|=bfr[5];
	status<<=8;
	status|=bfr[4];
	if(status)return -1;

	return 0;
}

int amtkvm(struct ctx *ctx,int persist,char *user,char *pass,int verbose)
{
	struct redirinfo info;

	switch(start_redirect(ctx,persist,"KVMR",&info))
	{
	case -1:if(verbose)fprintf(stderr,"connection/protocol error\n");
		return -1;

	case -2:if(verbose)fprintf(stderr,"busy, other client active, "
			"address: %s\n",inet_ntoa(info.addr));
		end_redirect(ctx);
		return -1;

	case 0:	if(verbose)fprintf(stderr,"redirect info: protocol %d.%d, "
			"firmware %d.%d, iana enterprise number %d\n",
			info.major,info.minor,info.fwmajor,info.fwminor,
			info.iana);
	}

	switch(authenticate(1,user,pass,io,ctx))
	{
	case -1:if(verbose)fprintf(stderr,
			"parameter/connection/protocol error");
		end_redirect(ctx);
		return -1;

	case -2:if(verbose)fprintf(stderr,"authentication failure\n");
		end_redirect(ctx);
		return -1;

	case 0:	if(verbose)fprintf(stderr,"successful authentication\n");
	}

	if(start_kvm(ctx))
	{
		if(verbose)fprintf(stderr,"connection/protocol error\n");
		end_redirect(ctx);
		return -1;
	}

	if(verbose)fprintf(stderr,"amt kvm (vnc) connection established.\n");
	return 0;
}
