/*
 *  AMT KVM enabled transparent proxy functions.
 *
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <pthread.h>
#include <poll.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <pwd.h>
#include <signal.h>
#include <sys/utsname.h>
#include "net.h"
#include "amtkvm.h"

#undef _POSIX_SOURCE
#define _LINUX_FS_H
#include <sys/syscall.h>
#include <linux/capability.h>
#include <sys/prctl.h>

extern int capset(cap_user_header_t header, const cap_user_data_t data);

#define CFGFILE "/etc/amtproxy.conf"

struct conndata
{
	int fd;
	union a4a6 src;
	union a4a6 dst;
};

struct redirect
{
	struct redirect *next;
	union a4a6 orig;
	union a4a6 real;
	char *user;
	char *pass;
	char *cacert;
};

static int mark=1;
static int timeout=10;
static int verbose=0;
static int persist=0;
static int dump=0;
static struct redirect *redir=NULL;

static int drop(char *name)
{
	struct __user_cap_header_struct h;
	struct __user_cap_data_struct c[2];
	struct passwd *pw;

	if(!(pw=getpwnam(name)))
	{
		perror("getpwnam");
		return -1;
	}

	if(setresgid(pw->pw_gid,pw->pw_gid,pw->pw_gid))
	{
		perror("setresgid");
		return -1;
	}

	memset(&h,0,sizeof(h));
	h.version=_LINUX_CAPABILITY_VERSION_2;
	memset(&c,0,sizeof(c));
	c[0].effective=1<<CAP_NET_ADMIN|1<<CAP_SETUID;
	c[0].permitted=1<<CAP_NET_ADMIN|1<<CAP_SETUID;
	capset(&h,c);
	prctl(PR_SET_KEEPCAPS,1,0,0,0);

	if(setresuid(pw->pw_uid,pw->pw_uid,pw->pw_uid))
	{
		perror("setresuid");
		return -1;
	}

	memset(&h,0,sizeof(h));
	h.version=_LINUX_CAPABILITY_VERSION;
	memset(&c,0,sizeof(c));
	c[0].effective=1<<CAP_NET_ADMIN;
	c[0].permitted=1<<CAP_NET_ADMIN;
	capset(&h,c);
	prctl(PR_SET_KEEPCAPS,0,0,0,0);

	return 0;
}

static void dropall(void)
{
	struct __user_cap_header_struct h;
	struct __user_cap_data_struct c[2];

	memset(&h,0,sizeof(h));
	h.version=_LINUX_CAPABILITY_VERSION_2;
	h.pid=0;
	memset(&c,0,sizeof(c));
	capset(&h,c);
}

static int precheck(void)
{
	int major;
	int minor;
	char *ptr;
	struct utsname bfr;

	if(geteuid())
	{
		fprintf(stderr,"must be root\n");
		return 1;
	}
	else if(getuid())
	{
		if(setuid(0))
		{
			perror("setuid");
			return 1;
		}
	}

	if(uname(&bfr))
	{
		perror("uname");
		return 1;
	}

	for(ptr=bfr.release;*ptr>='0'&&*ptr<='9';ptr++);
	if(*ptr)ptr++;
	major=atoi(bfr.release);
	minor=atoi(ptr);

	if(major<3||(major==3&&minor<7))
	{
		fprintf(stderr,"kernel too old, need at least 3.7\n");
		return 1;
	}

	return 0;
}

static void *connhandler(void *data)
{
	int srcport;
	int dstport;
	struct redirect *rdr;
	struct ctx *ctx;
	struct conndata *c=(struct conndata *)data;
	char srcaddr[HOSTLEN];
	char dstaddr[HOSTLEN];

	for(rdr=redir;rdr;rdr=rdr->next)if(!addrcmp(&c->dst,&rdr->orig,0))break;
	if(rdr)
	{
		if(verbose)
		{
			addr2str(&c->dst,srcaddr,&srcport);
			addr2str(&rdr->real,dstaddr,&dstport);
			printf("connection to %s port %d redirected to %s "
				"port %d%s\n",srcaddr,srcport,dstaddr,dstport,
				rdr->user?" (amt mode)":"");
		}
		c->dst=rdr->real;
	}

	if(!(ctx=doconnect(&c->dst,mark,timeout,rdr?rdr->cacert:NULL)))
	{
		if(verbose)
		{
			addr2str(&c->src,srcaddr,&srcport);
			addr2str(&c->dst,dstaddr,&dstport);
			printf("connection from %s port %d to %s port %d "
				"failed\n",srcaddr,srcport,dstaddr,dstport);
		}
		goto err1;
	}

	dropall();

	if(rdr)if(rdr->user)if(amtkvm(ctx,persist,rdr->user,rdr->pass,verbose))
	{
		if(verbose)
		{
			addr2str(&c->dst,dstaddr,&dstport);
			fprintf(stderr,"amt logon failure for %s port %d\n",
				dstaddr,dstport);
		}
		goto err2;
	}

	if(!(dotransfer(c->fd,ctx,dump)))
	{
		if(verbose)
		{
			addr2str(&c->src,srcaddr,&srcport);
			addr2str(&c->dst,dstaddr,&dstport);
			printf("data transfer between %s port %d and %s port "
				"%d complete\n",srcaddr,srcport,dstaddr,
				dstport);
		}
	}
	else if(verbose)
	{
		addr2str(&c->src,srcaddr,&srcport);
		addr2str(&c->dst,dstaddr,&dstport);
		printf("data transfer between %s port %d and %s port %d "
			"aborted\n",srcaddr,srcport,dstaddr,dstport);
	}

err2:	doclose(ctx);
err1:	close(c->fd);
	free(c);
	pthread_exit(NULL);
}

static int parse(char *config)
{
	int r=-1;
	int line;
	FILE *fp;
	char *srcaddr;
	char *srcport;
	char *dstaddr;
	char *dstport;
	char *amtuser;
	char *amtpass;
	char *cacert;
	struct redirect *e;
	struct redirect rdr;
	struct stat stb;
	char bfr[1024];

	if(!(fp=fopen(config,"re")))
	{
		perror("fopen");
		return 1;
	}

	if(fstat(fileno(fp),&stb))
	{
		perror("fstat");
		goto err;
	}

	if(!S_ISREG(stb.st_mode))
	{
		fprintf(stderr,"config must be a regular file\n");
		goto err;
	}

	if(((stb.st_mode&07777)&(~0700))||stb.st_uid)
	{
		fprintf(stderr,"bad config owner/permissions, refused\n");
		goto err;
	}

	for(line=0;fgets(bfr,sizeof(bfr),fp);line++)
	{
		srcaddr=strtok(bfr," \t\r\n");
		srcport=strtok(NULL," \t\r\n");
		dstaddr=strtok(NULL," \t\r\n");
		dstport=strtok(NULL," \t\r\n");
		amtuser=strtok(NULL," \t\r\n");
		amtpass=strtok(NULL," \t\r\n");
		cacert=strtok(NULL," \t\r\n");

		if(!srcaddr)continue;
		if(*srcaddr=='#')continue;

		if(!srcport||!dstaddr||!dstport||!*srcaddr||
			!*srcport||!*dstaddr||!*dstport)goto fail;

		if((amtuser&&!amtpass)||(!amtuser&&amtpass))goto fail;
		if(amtuser)if(!*amtuser||!*amtpass)goto fail;

		memset(&rdr,0,sizeof(rdr));
		if(inet_xton(srcaddr,srcport,&rdr.orig))goto fail;
		if(inet_xton(dstaddr,dstport,&rdr.real))goto fail;

		if(amtuser)
		{
			if(!(rdr.user=strdup(amtuser)))
			{
				perror("strdup");
				goto err;
			}

			if(!(rdr.pass=strdup(amtpass)))
			{
				perror("strdup");
				goto err;
			}

			if(cacert)
			{
				if(!(rdr.cacert=strdup(cacert)))
				{
					perror("strdup");
					goto err;
				}
			}
		}

		if(!(e=malloc(sizeof(struct redirect))))
		{
			perror("malloc");
			goto err;
		}

		*e=rdr;
		e->next=redir;
		redir=e;
	}

	r=0;

	while(0)
	{
fail:		fprintf(stderr,"configuration error in line %d\n",line);
	}
err:	fclose(fp);
	return r;
}

static void usage(void)
{
  fprintf(stderr,
  "Usage: amtproxy <options>\n"
  "\n"
  "Options are:\n"
  "\n"
  "-c <configuration-file>           Configuration file (" CFGFILE ")\n"
  "-u <user-to-run-as>               Valid user on system, mandatory\n"
  "-p <listening-port>               Valid Range: 1-65535, no default\n"
  "-m <netfilter-mark-to-set>        Valid Range: 1-MAXINT, default 1\n"
  "-t <connect-timeout-in-seconds>   Valid Range: 1-60, default 10\n"
  "-F <pid-file>                     File to write the pid to, default off\n"
  "-v                                Verbose output, default off\n"
  "-P                                Set AMT link preference bit, default off\n"
  "-f                                Stay in foreground, default off\n"
  "-d                                Dump proxied data to stdout, default off\n"
  "-h                                This help text\n");
  exit(1);
}

int main(int argc,char *argv[])
{
	int x;
	int port=0;
	int fg=0;
	int srcport;
	int dstport;
	FILE *fp;
	char *config=CFGFILE;
	char *user=NULL;
	char *pidfile=NULL;
	char srcaddr[HOSTLEN];
	char dstaddr[HOSTLEN];
	struct pollfd p;
	struct conndata c;
	struct conndata *cc;
	pthread_attr_t attr;
	pthread_t unused;

	while((x=getopt(argc,argv,"c:u:p:m:t:F:vPdfh"))!=-1)switch(x)
	{
	case 'c':
		config=optarg;
		break;

	case 'u':
		user=optarg;
		break;

	case 'p':
		port=atoi(optarg);
		break;

	case 'm':
		mark=(int)strtoul(optarg,NULL,0);
		break;

	case 't':
		timeout=atoi(optarg);
		break;

	case 'F':
		pidfile=optarg;
		break;

	case 'v':
		verbose=1;
		break;

	case 'P':
		persist=1;
		break;

	case 'd':
		dump=1;
	case 'f':
		fg=1;
		break;

	default:usage();
	}

	signal(SIGPIPE,SIG_IGN);

	if(!mark||!timeout||timeout>30||port<1||port>65535||!user)usage();

	if(precheck())return 1;

	if(pthread_attr_init(&attr))
	{
		perror("pthread_attr_init");
		return 1;
	}

	if(pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED))
	{
		perror("pthread_attr_setdetachstate");
		return 1;
	}

	if(netinit())return 1;

	p.events=POLLIN;
	if((p.fd=mklisten(port))==-1)
	{
		if(verbose)printf("listening port setup failed\n");
		return 1;
	}

	if(parse(config))return 1;

	if(!fg)if(daemon(0,0))
	{
		perror("daemon");
		return 1;
	}

	if(drop(user))
	{
		if(verbose)printf("dropping privileges failed\n");
		return 1;
	}

	if(pidfile)
	{
		if((fp=fopen(pidfile,"we")))
		{
			fprintf(fp,"%d\n",getpid());
			fclose(fp);
		}
		else if(verbose)printf("warning: can't write pid to %s\n",
			pidfile);
	}

	while(poll(&p,1,-1)==1)
	{
		if((c.fd=doaccept(p.fd,&c.src,&c.dst))==-1)continue;

		if(verbose)
		{
			addr2str(&c.src,srcaddr,&srcport);
			addr2str(&c.dst,dstaddr,&dstport);
			printf("connection request from %s port %d to "
				"%s port %d\n",srcaddr,srcport,dstaddr,dstport);
		}

		if(!addrcmp(&c.src,&c.dst,port))
		{
			if(verbose)
			{
				printf("loop detected for request to %s port %d"
					", closing connection\n",dstaddr,
					dstport);
			}
			close(c.fd);
			continue;
		}

		if(!(cc=malloc(sizeof(struct conndata))))
		{
			perror("malloc");
			close(c.fd);
			continue;
		}

		*cc=c;

		if(pthread_create(&unused,&attr,connhandler,cc))
		{
			perror("pthread_create");
			free(cc);
			close(c.fd);
		}
	}

	close(p.fd);
	netexit();
	return 0;
}
