/*
 *  Networking helper functions.
 *
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <linux/netfilter_ipv6/ip6_tables.h>
#include <sys/time.h>
#include <time.h>
#include <poll.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include "net.h"
#include "auth.h"

#if defined(USE_OPENSSL)
#include <openssl/ssl.h>
#include <openssl/err.h>
#elif defined(USE_GNUTLS)
#include <gnutls/gnutls.h>
#endif

#define CHUNK  16384

#define ERR(a,b) { perror(a) ; goto b; }

struct ctx
{
	int fd;
#if defined(USE_OPENSSL)
	SSL *ssl;
	SSL_CTX *ctx;
#elif defined(USE_GNUTLS)
	gnutls_session_t ssl;
	gnutls_certificate_credentials_t cred;
#endif
};

#if defined(USE_OPENSSL)

#define isssl(a) ((a)->ssl)

static int sslinit(struct ctx *ctx,char *cacert)
{
	int r;
	int c=0;

	if(!ctx)return 0;

	SSL_load_error_strings();
	SSL_library_init();

	if(!(ctx->ctx=SSL_CTX_new(SSLv23_client_method())))
	{
		ERR_print_errors_fp(stderr);
		goto err1;
	}

#if OPENSSL_VERSION_NUMBER >= 0x1000100FL
	SSL_CTX_set_options(ctx->ctx,
		SSL_OP_NO_TLSv1_1|SSL_OP_NO_TLSv1_2);
#endif

	if(!SSL_CTX_load_verify_locations(ctx->ctx,cacert,NULL))
	{
		ERR_print_errors_fp(stderr);
		goto err2;
	}

	SSL_CTX_set_verify_depth(ctx->ctx,5);
	SSL_CTX_set_verify(ctx->ctx,SSL_VERIFY_PEER,NULL);

	if(!(ctx->ssl=SSL_new(ctx->ctx)))
	{
		ERR_print_errors_fp(stderr);
		goto err2;
	}

	if(!SSL_set_fd(ctx->ssl,ctx->fd))
	{
		ERR_print_errors_fp(stderr);
		goto err3;
	}

repeat:	if((r=SSL_connect(ctx->ssl))!=1)
	{
		switch(SSL_get_error(ctx->ssl,r))
		{
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_WRITE:
			if(++c<100)
			{
				usleep(10000);
				goto repeat;
			}
		}
		ERR_print_errors_fp(stderr);
		goto err3;
	}

	return 0;

err3:	SSL_free(ctx->ssl);
err2:	SSL_CTX_free(ctx->ctx);
err1:	return -1;
}

static void sslexit(struct ctx *ctx)
{
	if(ctx)if(ctx->ssl)
	{
		SSL_shutdown(ctx->ssl);
		SSL_free(ctx->ssl);
		SSL_CTX_free(ctx->ctx);
	}
}

static int sslpoll(SSL *ssl,struct pollfd *p,int n,int timeout)
{
	int r;

	if(ssl)if(SSL_pending(ssl))if(p[0].events&POLLIN)
	{
		if((r=poll(p,n,0))==-1)return -1;
		if(!p[0].revents)r++;
		p[0].revents|=POLLIN;
		return r;
	}
	return poll(p,n,timeout);
}

static ssize_t doread(struct ctx *ctx,void *buf,size_t count)
{
	int l;
	int c=0;

	if(!ctx->ssl)return read(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=SSL_read(ctx->ssl,buf,count))>0)return l;

	switch(SSL_get_error(ctx->ssl,l))
	{
	case SSL_ERROR_WANT_READ:
	case SSL_ERROR_WANT_WRITE:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
		break;
	case SSL_ERROR_WANT_X509_LOOKUP:
		return -1;
	case SSL_ERROR_ZERO_RETURN:
		return 0;
	case SSL_ERROR_SSL:
		ERR_print_errors_fp(stderr);
	}

	return -1;
}

static ssize_t dowrite(struct ctx *ctx,void *buf,size_t count)
{
	int l;
	int c=0;

	if(!ctx->ssl)return write(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=SSL_write(ctx->ssl,buf,count))>0)return l;

	switch(SSL_get_error(ctx->ssl,l))
	{
	case SSL_ERROR_WANT_READ:
	case SSL_ERROR_WANT_WRITE:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
		break;
	case SSL_ERROR_WANT_X509_LOOKUP:
		return -1;
	case SSL_ERROR_ZERO_RETURN:
		return 0;
	case SSL_ERROR_SSL:
		ERR_print_errors_fp(stderr);
	}

	return -1;
}

#elif defined(USE_GNUTLS)

#define isssl(a) ((a)->ssl)

static int vrycb(gnutls_session_t ssl)
{
	int r;
	int type;
	unsigned int status;
	gnutls_datum_t msg;

	if((r=gnutls_certificate_verify_peers3(ssl,NULL,&status))<0)
	{
		fprintf(stderr,"gnutls_certificate_verify_peers3: %s\n",
			gnutls_strerror(r));
		return GNUTLS_E_CERTIFICATE_ERROR;
	}

	if(status)
	{
		type=gnutls_certificate_type_get(ssl);
		if((r=gnutls_certificate_verification_status_print(status,type,
			&msg,0))<0)
		{
			fprintf(stderr,"gnutls_certificate_verification_"
				"status_print %s\n",gnutls_strerror(r));
		}
		else
		{
			fprintf(stderr,"certificate status: %s\n",msg.data);
			gnutls_free(msg.data);
		}
		return GNUTLS_E_CERTIFICATE_ERROR;
	}

	return 0;
}

static int sslinit(struct ctx *ctx,char *cacert)
{
	int r;
	const char *e;

	if(!ctx)
	{
		if((r=gnutls_global_init()))
		{
			fprintf(stderr,"gnutls_global_init: %s\n",
				gnutls_strerror(r));
			goto err1;
		}

		return 0;
	}

	if((r=gnutls_certificate_allocate_credentials(&ctx->cred)))
	{
		fprintf(stderr,"gnutls_certificate_allocate_credentials: "
			"%s\n",gnutls_strerror(r));
		goto err1;
	}

	if((r=gnutls_certificate_set_x509_trust_file(ctx->cred,cacert,
		GNUTLS_X509_FMT_PEM))<0)
	{
		fprintf(stderr,"gnutls_certificate_set_x509_trust_file: "
			"%s\n",gnutls_strerror(r));
		goto err2;
	}

	gnutls_certificate_set_verify_function(ctx->cred,vrycb);

	if((r=gnutls_init(&ctx->ssl,GNUTLS_CLIENT)))
	{
		fprintf(stderr,"gnutls_init: %s\n",gnutls_strerror(r));
		goto err2;
	}

	/* oh well, isn't _that_ easy ?!? :-(  ... compare to openssl ... */
	if((r=gnutls_priority_set_direct(ctx->ssl,"NONE:+AES-256-CBC:"
		"+AES-128-CBC:+3DES-CBC:+COMP-NULL:+CTYPE-X509:+VERS-SSL3.0:"
		"+SHA256:+SHA1:+RSA:%UNSAFE_RENEGOTIATION",&e)))
	{
		fprintf(stderr,"gnutls_priority_set_direct: %s\n",
			gnutls_strerror(r));
		if(r==GNUTLS_E_INVALID_REQUEST)
			fprintf(stderr,"additional info: %s\n",e);
		goto err3;
	}

	if((r=gnutls_credentials_set(ctx->ssl,GNUTLS_CRD_CERTIFICATE,
		ctx->cred)))
	{
		fprintf(stderr,"gnutls_credentials_set: %s\n",
			gnutls_strerror(r));
		goto err3;
	}

	gnutls_transport_set_int(ctx->ssl,ctx->fd);

	gnutls_handshake_set_timeout(ctx->ssl,GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);

	do
	{
		r=gnutls_handshake(ctx->ssl);
	} while(r<0&&!gnutls_error_is_fatal(r));
	if(r<0)
	{
		fprintf(stderr,"gnutls_handshake: %s\n",gnutls_strerror(r));
		goto err3;
	}

	return 0;

err3:	gnutls_deinit(ctx->ssl);
err2:	gnutls_certificate_free_credentials(ctx->cred);
err1:	return -1;
}

static void sslexit(struct ctx *ctx)
{
	if(!ctx)gnutls_global_deinit();
	else if(ctx->ssl)
	{
		gnutls_deinit(ctx->ssl);
		gnutls_certificate_free_credentials(ctx->cred);
	}
}

static int sslpoll(gnutls_session_t ssl,struct pollfd *p,int n,int timeout)
{
	int r;

	if(ssl)if(gnutls_record_check_pending(ssl))if(p[0].events&POLLIN)
	{
		if((r=poll(p,n,0))==-1)return -1;
		if(!p[0].revents)r++;
		p[0].revents|=POLLIN;
		return r;
	}
	return poll(p,n,timeout);
}

static ssize_t doread(struct ctx *ctx,void *buf,size_t count)
{
	ssize_t l;
	int c=0;
	int r;

	if(!ctx->ssl)return read(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=gnutls_record_recv(ctx->ssl,buf,count))>0)return l;

	switch(l)
	{
	case GNUTLS_E_REHANDSHAKE:
		do
		{
			r=gnutls_handshake(ctx->ssl);
		} while(r<0&&!gnutls_error_is_fatal(r));
		if(r<0)
		{
			fprintf(stderr,"gnutls_handshake: %s\n",
			gnutls_strerror(r));
			return -1;
		}
	case GNUTLS_E_INTERRUPTED:
	case GNUTLS_E_AGAIN:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
	default:fprintf(stderr,"gnutls_record_recv: %s\n",gnutls_strerror(l));
	case GNUTLS_E_PUSH_ERROR:
	case GNUTLS_E_PULL_ERROR:
		return -1;
	}
}

static ssize_t dowrite(struct ctx *ctx,void *buf,size_t count)
{
	ssize_t l;
	int c=0;
	int r;

	if(!ctx->ssl)return write(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=gnutls_record_send(ctx->ssl,buf,count))>0)return l;

	switch(l)
	{
	case GNUTLS_E_REHANDSHAKE:
		do
		{
			r=gnutls_handshake(ctx->ssl);
		} while(r<0&&!gnutls_error_is_fatal(r));
		if(r<0)
		{
			fprintf(stderr,"gnutls_handshake: %s\n",
			gnutls_strerror(r));
			return -1;
		}
	case GNUTLS_E_INTERRUPTED:
	case GNUTLS_E_AGAIN:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
	default:fprintf(stderr,"gnutls_record_send: %s\n",gnutls_strerror(l));
	case GNUTLS_E_PUSH_ERROR:
	case GNUTLS_E_PULL_ERROR:
		return -1;
	}
}

#else

#define sslinit(a,b)		0
#define sslexit(a)
#define sslpoll(a,b,c,d)	poll(b,c,d)
#define doread(a,b,c)		read(((struct ctx *)a)->fd,b,c)
#define dowrite(a,b,c)		write(((struct ctx *)a)->fd,b,c)
#define isssl(a)		0

#endif

static void dumper(unsigned char *data,int len,char *dir)
{
	int off;
	int i;
	int j;
	struct timeval tv;
	struct tm tm;
	char stamp[64];
	char hex[50];
	char chr[17];

	gettimeofday(&tv,NULL);
	localtime_r(&tv.tv_sec,&tm);
	strftime(stamp,sizeof(stamp),"%F %T",&tm);

	printf("%s.%06ld %4d bytes %s\n",stamp,tv.tv_usec,len,dir);

	off=0;
	while(len)
	{
		for(j=0,i=0;i<len&&i<16;i++)
		{
			hex[j]=data[i]>>4;
			if(hex[j]<10)hex[j++]+='0';
			else hex[j++]+='a'-10;

			hex[j]=data[i]&0xf;
			if(hex[j]<10)hex[j++]+='0';
			else hex[j++]+='a'-10;

			hex[j++]=' ';
			if(i==7)hex[j++]=' ';

			if(data[i]>=0x20&&data[i]<=0x7e)chr[i]=data[i];
			else chr[i]='.';
		}

		for(;i<16;i++)
		{
			hex[j++]=' ';
			hex[j++]=' ';
			hex[j++]=' ';
			if(i==7)hex[j++]=' ';
			chr[i]=' ';
		}

		hex[j]=0;
		chr[i]=0;

		printf("  0x%04x  %s %s\n",off,hex,chr);

		if(len<16)break;
		off+=16;
		data+=16;
		len-=16;
	}

	fflush(stdout);
}

static int slowtransfer(int s1,struct ctx *ctx,int dump)
{
	int len;
	int fill1=0;
	int fill2=0;
	int closed1in=0;
	int closed2in=0;
	int closed1out=0;
	int closed2out=0;
	unsigned char b1[CHUNK];
	unsigned char b2[CHUNK];
	struct pollfd p[2];

	p[0].fd=ctx->fd;
	p[1].fd=s1;

	while(!closed1out||!closed2out)
	{
		p[0].events=0;
		if(fill1<sizeof(b1)&&!closed1in)p[0].events|=POLLIN;
		if(fill2&&!closed1out)p[0].events|=POLLOUT;

		p[1].events=0;
		if(fill2<sizeof(b2)&&!closed2in)p[1].events|=POLLIN;
		if(fill1&&!closed2out)p[1].events|=POLLOUT;

		if(p[0].events||p[1].events)while(sslpoll(ctx->ssl,p,2,-1)<=0);

		if(p[0].revents&POLLIN)
		{
			if((len=doread(ctx,b1+fill1,sizeof(b1)-fill1))<=0)
			{
				if(!closed1in)shutdown(p[0].fd,SHUT_RD);
				closed1in=1;
			}
			else
			{
				if(dump)dumper(b1+fill1,len,"dst->src");
				fill1+=len;
				if(closed2out)fill1=0;
			}
		}

		if(p[1].revents&POLLIN)
		{
			if((len=read(p[1].fd,b2+fill2,sizeof(b2)-fill2))<=0)
			{
				if(!closed2in)shutdown(p[1].fd,SHUT_RD);
				closed2in=1;
			}
			else
			{
				if(dump)dumper(b2+fill2,len,"src->dst");
				fill2+=len;
				if(closed1out)fill2=0;
			}
		}

		if(p[0].revents&(POLLERR|POLLHUP))
		{
			if(!closed2in)shutdown(p[1].fd,SHUT_RD);
			closed2in=1;
			if(!closed1out)shutdown(p[0].fd,SHUT_WR);
			closed1out=1;
			fill2=0;
		}

		if(p[1].revents&(POLLERR|POLLHUP))
		{
			if(!closed1in)shutdown(p[0].fd,SHUT_RD);
			closed1in=1;
			if(!closed2out)shutdown(p[1].fd,SHUT_WR);
			closed2out=1;
			fill1=0;
		}

		if(fill2)if(p[0].revents&POLLOUT)
		{
			if((len=dowrite(ctx,b2,fill2))<=0)
			{
				if(!closed2in)shutdown(p[1].fd,SHUT_RD);
				closed2in=1;
				if(!closed1out)shutdown(p[0].fd,SHUT_WR);
				closed1out=1;
				fill2=0;
			}
			else
			{
				if(fill2!=len)memmove(b2,b2+len,fill2-len);
				fill2-=len;
			}
		}

		if(fill1)if(p[1].revents&POLLOUT)
		{
			if((len=write(p[1].fd,b1,fill1))<=0)
			{
				if(!closed1in)shutdown(p[0].fd,SHUT_RD);
				closed1in=1;
				if(!closed2out)shutdown(p[1].fd,SHUT_WR);
				closed2out=1;
				fill1=0;
			}
			else
			{
				if(fill1!=len)memmove(b1,b1+len,fill1-len);
				fill1-=len;
			}
		}

		if(closed1out)fill2=0;
		if(closed2out)fill1=0;
		if(closed1in&&!fill1)
		{
			if(!closed2out)shutdown(p[1].fd,SHUT_WR);
			closed2out=1;
		}
		if(closed2in&&!fill2)
		{
			if(!closed1out)shutdown(p[0].fd,SHUT_WR);
			closed1out=1;
		}
	}

	return 0;
}

static int fasttransfer(int s1,int s2)
{
	int r=-1;
	int p1[2];
	int p2[2];
	int e[4];
	struct pollfd p[4];
	struct pollfd n[2];

	if(pipe2(p1,O_NONBLOCK|O_CLOEXEC))ERR("pipe",err1);
	if(pipe2(p2,O_NONBLOCK|O_CLOEXEC))ERR("pipe",err2);

	p[0].fd=p1[0];
	p[0].events=POLLIN;
	p[1].fd=p1[1];
	p[1].events=POLLOUT;
	p[2].fd=p2[0];
	p[2].events=POLLIN;
	p[3].fd=p2[1];
	p[3].events=POLLOUT;
	n[0].fd=s1;
	n[1].fd=s2;
	e[0]=e[2]=POLLOUT;
	e[1]=e[3]=POLLIN|POLLRDHUP;

	while(1)
	{
		if(poll(p,4,0)==-1)ERR("poll",err3);

		if(!e[1]&&!(p[0].revents&POLLIN))if(e[0])
		{
			e[0]=0;
			shutdown(n[1].fd,SHUT_WR);
		}

		if(!e[3]&&!(p[2].revents&POLLIN))if(e[2])
		{
			e[2]=0;
			shutdown(n[0].fd,SHUT_WR);
		}

		n[0].events=n[1].events=0;
		if(p[0].revents&POLLIN)n[1].events|=e[0];
		if(p[1].revents&POLLOUT)n[0].events|=e[1];
		if(p[2].revents&POLLIN)n[0].events|=e[2];
		if(p[3].revents&POLLOUT)n[1].events|=e[3];

		if(!n[0].events&&!n[1].events)break;

		if(poll(n,2,-1)==-1)ERR("poll",err3);

		if(n[0].revents&POLLRDHUP)
		{
			e[1]=0;
			shutdown(n[0].fd,SHUT_RD);
		}
		else if(n[0].revents&POLLIN)
		{
			switch(splice(n[0].fd,NULL,p[1].fd,NULL,CHUNK,
				SPLICE_F_NONBLOCK))
			{
			case -1:ERR("splice",err3);
			case 0:	e[1]=0;
				shutdown(n[0].fd,SHUT_RD);
			}
		}

		if(n[0].revents&(POLLERR|POLLHUP))
		{
			e[2]=0;
			e[3]=0;
			shutdown(n[0].fd,SHUT_WR);
			shutdown(n[1].fd,SHUT_RD);
		}
		else if(n[0].revents&POLLOUT)
		{
			switch(splice(p[2].fd,NULL,n[0].fd,NULL,CHUNK,
				SPLICE_F_NONBLOCK))
			{
			case -1:ERR("splice",err3);
			case 0:	e[2]=0;
				e[3]=0;
				shutdown(n[0].fd,SHUT_WR);
				shutdown(n[1].fd,SHUT_RD);
			}
		}

		if(n[1].revents&POLLRDHUP)
		{
			e[3]=0;
			shutdown(n[1].fd,SHUT_RD);
		}
		else if(n[1].revents&POLLIN)
		{
			switch(splice(n[1].fd,NULL,p[3].fd,NULL,CHUNK,
				SPLICE_F_NONBLOCK))
			{
			case -1:ERR("splice",err3);
			case 0:	e[3]=0;
				shutdown(n[1].fd,SHUT_RD);
			}
		}

		if(n[1].revents&(POLLERR|POLLHUP))
		{
			e[0]=0;
			e[1]=0;
			shutdown(n[0].fd,SHUT_RD);
			shutdown(n[1].fd,SHUT_WR);
		}
		else if(n[1].revents&POLLOUT)
		{
			switch(splice(p[0].fd,NULL,n[1].fd,NULL,CHUNK,
				SPLICE_F_NONBLOCK))
			{
			case -1:ERR("splice",err3);
			case 0:	e[0]=0;
				e[1]=0;
				shutdown(n[0].fd,SHUT_RD);
				shutdown(n[1].fd,SHUT_WR);
			}
		}
	}

	r=0;

err3:	close(p2[0]);
	close(p2[1]);
err2:	close(p1[0]);
	close(p1[1]);
err1:	return r;
}

static int dnslookup(union a4a6 *req,union a4a6 *res)
{
	int r;
	struct addrinfo *ptr;
	struct addrinfo *list=NULL;
	struct addrinfo hints;
	union a4a6 a4;
	union a4a6 a6;

	if(req->dns.family!=AF_UNSPEC)return -1;

	memset(&a4,0,sizeof(a4));
	memset(&a6,0,sizeof(a6));

	memset(&hints,0,sizeof(hints));
	hints.ai_family=AF_UNSPEC;
	hints.ai_flags=AI_ADDRCONFIG|AI_IDN;

	r=getaddrinfo(req->dns.host+(*req->dns.host=='!'?1:0),NULL,&hints,
		&list);

	if(!r&&list)
	{
		memset(res,0,sizeof(union a4a6));

		for(ptr=list;ptr&&r!=3;ptr=ptr->ai_next)switch(ptr->ai_family)
		{
		case AF_INET:
			if(r&1)break;
			a4.a4.sin_family=AF_INET;
			a4.a4.sin_addr=
				((struct sockaddr_in *)ptr->ai_addr)->sin_addr;
			a4.a4.sin_port=htons(req->dns.port);
			r|=1;
			break;

		case AF_INET6:
			if(r&2)break;
			a6.a6.sin6_family=AF_INET6;
			memcpy(a6.a6.sin6_addr.s6_addr,
				((struct sockaddr_in6 *)
				ptr->ai_addr)->sin6_addr.s6_addr,16);
			a6.a6.sin6_port=htons(req->dns.port);
			r|=2;
			break;
		}

		freeaddrinfo(list);

		if(*req->dns.host=='!')
		{
			if(r&1)
			{
				res[0]=a4;
				if(r&2)res[1]=a6;
			}
			else if(r&2)res[0]=a6;
		}
		else
		{
			if(r&2)
			{
				res[0]=a6;
				if(r&1)res[1]=a4;
			}
			else if(r&1)res[0]=a4;
		}

		switch(r)
		{
		case 3:	r=2;
		case 0:	break;
		default:r=1;
		}
	}

	return r;
}

int io(void *parm,unsigned char *data,int len,int mode)
{
	int l;
	struct pollfd p;

	p.fd=((struct ctx *)parm)->fd;
	p.events=POLLIN;

	switch(mode)
	{
	case READ:
		for(l=0;len;data+=l,len-=l)
repeat1:		switch(sslpoll(((struct ctx *)parm)->ssl,&p,1,20000))
		{
		case -1:goto repeat1;
		case 0:	return -1;
		case 1:	if(!(p.revents&POLLIN))return -1;
			if((l=doread(parm,data,len))<=0)return -1;
		}
		return 0;

	case WRITE:
		for(p.events=POLLOUT,l=0;len;data+=l,len-=l)
repeat2:		switch(poll(&p,1,20000))
		{
		case -1:goto repeat2;
		case 0:	return -1;
		case 1:	if(!(p.revents&POLLOUT))return -1;
			if((l=dowrite(parm,data,len))<=0)return -1;
		}
		return 0;

	case RANDOM:
		if((p.fd=open("/dev/urandom",O_RDONLY))==-1)return -1;
		l=read(p.fd,data,len);
		close(p.fd);
		return l==len?0:-1;
	}

	return -1;
}

int mklisten(int port)
{
	int s;
	int x;
	struct linger l;
	struct sockaddr_in6 a;

	if(port<1||port>65535)goto err1;

	memset(&a,0,sizeof(a));
	a.sin6_family=AF_INET6;
	a.sin6_addr=in6addr_any;
	a.sin6_port=htons(port);

	if((s=socket(AF_INET6,SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_TCP))==-1)ERR("socket",err1);

	x=1;
	if(setsockopt(s,SOL_IP,IP_TRANSPARENT,&x,sizeof(x)))
		ERR("setsockopt IP_TRANSPARENT",err2);

	x=1;
	if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&x,sizeof(x)))
		ERR("setsockopt SO_REUSEADDR",err2);

	x=0;
	if(setsockopt(s,IPPROTO_IPV6,IPV6_V6ONLY,&x,sizeof(x)))
		ERR("setsockopt IPV6_V6ONLY",err2);

	if(bind(s,(struct sockaddr *)(&a),sizeof(a)))ERR("bind",err2);

	l.l_onoff=1;
	l.l_linger=10;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))
		ERR("setsockopt SO_LINGER",err2);

	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))
		ERR("setsockopt TCP_NODELAY",err2);

	if(listen(s,255))ERR("listen",err2);

	return s;

err2:	close(s);
err1:	return -1;
}

int doaccept(int fd,union a4a6 *src,union a4a6 *dst)
{
	int s;
	int x;
	socklen_t len;
	struct linger l;

	len=sizeof(union a4a6);
	if((s=accept4(fd,(struct sockaddr *)src,&len,
		SOCK_CLOEXEC|SOCK_NONBLOCK))==-1)ERR("accept4",err1);

	l.l_onoff=1;
	l.l_linger=10;
	if(setsockopt(s,SOL_SOCKET,SO_LINGER,&l,sizeof(l)))
		ERR("setsockopt SO_LINGER",err2);

	if(setsockopt(s,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))
		ERR("setsockopt TCP_NODELAY",err2);

	if(src->a4.sin_family==AF_INET)x=SOL_IP;
	else if(IN6_IS_ADDR_V4MAPPED(src->a6.sin6_addr.s6_addr))
	{
		dst->a4.sin_family=AF_INET;
		dst->a4.sin_port=src->a6.sin6_port;
		memcpy(&dst->a4.sin_addr.s_addr,src->a6.sin6_addr.s6_addr+12,4);
		src->a4=dst->a4;
		x=SOL_IP;
	}
	else x=IPPROTO_IPV6;

	len=sizeof(union a4a6);
	if(getsockopt(s,x,IP6T_SO_ORIGINAL_DST,(struct sockaddr *)dst,&len))
		ERR("getsockopt IP6T_SO_ORIGINAL_DST",err2);

	return s;

err2:	close(s);
err1:	return -1;
}

struct ctx *doconnect(union a4a6 *dst,int mark,int timeout,char *cacert)
{
	int x;
	socklen_t y;
	struct ctx *ctx;
	struct pollfd p;
	union a4a6 res[2];

	if(dst->dns.family==AF_UNSPEC)
	{
		if(dnslookup(dst,res)<=0)
		{
			fprintf(stderr,"dns lookup failure\n");
			goto err1;
		}
		dst=res;
	}

	if(!(ctx=malloc(sizeof(struct ctx))))ERR("malloc",err1);
	memset(ctx,0,sizeof(struct ctx));

	if((ctx->fd=socket(dst->a4.sin_family,
		SOCK_STREAM|SOCK_NONBLOCK|SOCK_CLOEXEC,
		IPPROTO_TCP))==-1)ERR("socket",err2);

	if(mark)if(setsockopt(ctx->fd,SOL_SOCKET,SO_MARK,&mark,sizeof(mark)))
		ERR("setsockopt SO_MARK",err3);

	if(connect(ctx->fd,(struct sockaddr *)(dst),sizeof(union a4a6)))
	{
		if(errno!=EINPROGRESS)ERR("connect",err3);

		p.fd=ctx->fd;
		p.events=POLLOUT;
		switch(poll(&p,1,timeout*1000))
		{
		case -1:ERR("poll",err3);
		case 0:	fprintf(stderr,"connect timeout\n");
			goto err3;
		}

		y=sizeof(x);
		if(getsockopt(ctx->fd,SOL_SOCKET,SO_ERROR,&x,&y))
			ERR("getsockopt SO_ERROR",err3);

		if(x)
		{
			fprintf(stderr,"connect error\n");
			goto err3;
		}
	}

	if(cacert)if(sslinit(ctx,cacert))goto err3;

	return ctx;

err3:	close(ctx->fd);
err2:	free(ctx);
err1:	return NULL;
}

void doclose(struct ctx *ctx)
{
	sslexit(ctx);
	close(ctx->fd);
	free(ctx);
}

int dotransfer(int fd,struct ctx *ctx,int dump)
{
	if(dump||isssl(ctx))return slowtransfer(fd,ctx,dump);
	else return fasttransfer(fd,ctx->fd);
}

int addrcmp(union a4a6 *src,union a4a6 *dst,int port)
{
	int n;
	union a4a6 res[2];

	if(dst->dns.family==AF_UNSPEC)
	{
		if((n=dnslookup(dst,res))<=0)return -1;
		dst=res;
	}
	else n=1;

	for(;n--;dst++)
	{
		if(src->a4.sin_family!=dst->a4.sin_family)continue;

		if(src->a4.sin_family==AF_INET)
		{
			if(src->a4.sin_addr.s_addr!=dst->a4.sin_addr.s_addr)
				continue;
			if(port)
			{
				if(dst->a4.sin_port!=htons(port))continue;
			}
			else if(src->a4.sin_port!=dst->a4.sin_port)continue;
			return 0;
		}

		if(memcmp(src->a6.sin6_addr.s6_addr,dst->a6.sin6_addr.s6_addr,
			16))continue;
		if(port)
		{
			if(dst->a6.sin6_port!=htons(port))continue;
		}
		else if(src->a6.sin6_port!=dst->a6.sin6_port)continue;
		return 0;
	}

	return -1;
}

void addr2str(union a4a6 *addr,char *str,int *port)
{
	if(addr->a4.sin_family==AF_INET)
	{
		inet_ntop(AF_INET,&addr->a4.sin_addr,str,INET6_ADDRSTRLEN);
		*port=ntohs(addr->a4.sin_port);
	}
	else if(addr->a6.sin6_family==AF_INET6)
	{
		inet_ntop(AF_INET6,&addr->a6.sin6_addr,str,INET6_ADDRSTRLEN);
		*port=ntohs(addr->a6.sin6_port);
	}
	else
	{
		strcpy(str,addr->dns.host+(*addr->dns.host=='!'?1:0));
		*port=addr->dns.port;
	}
}

int inet_xton(char *addr,char *port,union a4a6 *dst)
{
	int val;

	val=atoi(port);
	if(val<1||val>65535)return -1;

	switch(inet_pton(AF_INET,addr,&dst->a4.sin_addr.s_addr))
	{
	case 1:	dst->a4.sin_port=htons(val);
		dst->a4.sin_family=AF_INET;
		break;

	case -1:perror("inet_pton");
		return -1;

	case 0:	switch(inet_pton(AF_INET6,addr,
			dst->a6.sin6_addr.s6_addr))
		{
		case 1:	dst->a6.sin6_port=htons(val);
			dst->a6.sin6_family=AF_INET6;
			break;

		case -1:perror("inet_pton");
			return -1;

		case 0:	dst->dns.family=AF_UNSPEC;
			dst->dns.port=val;
			if(strlen(addr)>=sizeof(dst->dns.host))return -1;
			strcpy(dst->dns.host,addr);
			return 0;
		}
		break;
	}

	return 0;
}

int netinit(void)
{
	return sslinit(NULL,NULL);
}

void netexit(void)
{
	sslexit(NULL);
}
